/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the multiple precision addition and subtraction functions.
 *
 * @version $Id$
 * @ingroup bn
 */

#include <gmp.h>

#include "relic_core.h"
#include "relic_bn.h"
#include "relic_bn_low.h"
#include "relic_error.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void bn_gcd_ext_gmp(bn_t c, bn_t d, bn_t e, bn_t a, bn_t b) {
	bn_t u, v;
	mp_size_t cn, dn;

	bn_null(u);
	bn_null(v);

	TRY {
		bn_new(u);
		bn_new(v);

		bn_copy(u, a);
		bn_copy(v, b);
//		memset(c->dp + c->used, 0, (v->used - u->used) * sizeof(dig_t));
		cn = mpn_gcdext(c->dp, d->dp, &dn, u->dp, v->used, v->dp, v->used);
		memset(c->dp + cn, 0, (BN_SIZE - cn) * sizeof(dig_t));
		memset(d->dp + dn, 0, (BN_SIZE - dn) * sizeof(dig_t));
		c->used = cn;
		if (dn < 0) {
			d->used = -dn;
			d->sign = BN_NEG;
		} else {
			d->used = dn;
			d->sign = BN_POS;
		}
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(u);
		bn_free(v);
	}
}


#if BN_GCD == BASIC || !defined(STRIP)

void bn_gcd_basic(bn_t c, bn_t a, bn_t b) {
	bn_t u, v;

	bn_null(u);
	bn_null(v);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		return;
	}

	TRY {
		bn_new(u);
		bn_new(v);

		bn_abs(u, a);
		bn_abs(v, b);
		while (!bn_is_zero(v)) {
			bn_copy(c, v);
			bn_mod(v, u, v);
			bn_copy(u, c);
		}
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(u);
		bn_free(v);
	}
}

void bn_gcd_ext_basic(bn_t c, bn_t d, bn_t e, bn_t a, bn_t b) {
	bn_t u, v, x_1, y_1, q, r;

	bn_null(u);
	bn_null(v);
	bn_null(x_1);
	bn_null(y_1);
	bn_null(q);
	bn_null(r);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		bn_zero(d);
		bn_set_dig(e, 1);
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		bn_set_dig(d, 1);
		bn_zero(e);
		return;
	}

	TRY {
		bn_new(u);
		bn_new(v);
		bn_new(x_1);
		bn_new(y_1);
		bn_new(q);
		bn_new(r);

		bn_abs(u, a);
		bn_abs(v, b);

		bn_zero(x_1);
		bn_set_dig(y_1, 1);

		if (e != NULL) {
			bn_set_dig(d, 1);
			bn_zero(e);

			while (!bn_is_zero(v)) {
				bn_div_rem(q, r, u, v);

				bn_copy(u, v);
				bn_copy(v, r);

				bn_mul(c, q, x_1);
				bn_sub(r, d, c);
				bn_copy(d, x_1);
				bn_copy(x_1, r);

				bn_mul(c, q, y_1);
				bn_sub(r, e, c);
				bn_copy(e, y_1);
				bn_copy(y_1, r);
			}
		} else {
			bn_set_dig(d, 1);

			while (!bn_is_zero(v)) {
				bn_div_rem(q, r, u, v);

				bn_copy(u, v);
				bn_copy(v, r);

				bn_mul(c, q, x_1);
				bn_sub(r, d, c);
				bn_copy(d, x_1);
				bn_copy(x_1, r);
			}
		}
		bn_copy(c, u);
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(u);
		bn_free(v);
		bn_free(x_1);
		bn_free(y_1);
		bn_free(q);
		bn_free(r);
	}
}

#endif

#if BN_GCD == LEHME || !defined(STRIP)

void bn_gcd_lehme(bn_t c, bn_t a, bn_t b) {
	bn_t x, y, u, v, t0, t1, t2, t3;
	dig_t _x, _y, q, _q, t, _t;
	dsi_t _a, _b, _c, _d;

	bn_null(x);
	bn_null(y);
	bn_null(u);
	bn_null(v);
	bn_null(t0);
	bn_null(t1);
	bn_null(t2);
	bn_null(t3);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		return;
	}

	/*
	 * Taken from Handbook of Hyperelliptic and Elliptic Cryptography.
	 */
	TRY {
		bn_new(x);
		bn_new(y);
		bn_new(u);
		bn_new(v);
		bn_new(t0);
		bn_new(t1);
		bn_new(t2);
		bn_new(t3);

		if (bn_cmp(a, b) != CMP_LT) {
			bn_abs(x, a);
			bn_abs(y, b);
		} else {
			bn_abs(x, b);
			bn_abs(y, a);
		}
		while (y->used > 1) {
			bn_rsh(u, x, bn_bits(x) - BN_DIGIT);
			_x = u->dp[0];
			bn_rsh(v, y, bn_bits(x) - BN_DIGIT);
			_y = v->dp[0];
			_a = _d = 1;
			_b = _c = 0;
			t = 0;
			if (_y != 0) {
				q = _x / _y;
				t = _x % _y;
			}
			if (t >= ((dig_t)1 << (BN_DIGIT / 2))) {
				while (1) {
					_q = _y / t;
					_t = _y % t;
					if (_t < ((dig_t)1 << (BN_DIGIT / 2))) {
						break;
					}
					_x = _y;
					_y = t;
					t = _a - q * _c;
					_a = _c;
					_c = t;
					t = _b - q * _d;
					_b = _d;
					_d = t;
					t = _t;
					q = _q;
				}
			}
			if (_b == 0) {
				bn_mod(t0, x, y);
				bn_copy(x, y);
				bn_copy(y, t0);
			} else {
				bn_rsh(u, x, bn_bits(x) - 2 * BN_DIGIT);
				bn_rsh(v, y, bn_bits(x) - 2 * BN_DIGIT);
				if (_a < 0) {
					bn_mul_dig(t0, u, -_a);
					bn_neg(t0, t0);
				} else {
					bn_mul_dig(t0, u, _a);
				}
				if (_b < 0) {
					bn_mul_dig(t1, v, -_b);
					bn_neg(t1, t1);
				} else {
					bn_mul_dig(t1, v, _b);
				}
				if (_c < 0) {
					bn_mul_dig(t2, u, -_c);
					bn_neg(t2, t2);
				} else {
					bn_mul_dig(t2, u, _c);
				}
				if (_d < 0) {
					bn_mul_dig(t3, v, -_d);
					bn_neg(t3, t3);
				} else {
					bn_mul_dig(t3, v, _d);
				}
				bn_add(u, t0, t1);
				bn_add(v, t2, t3);
				bn_rsh(t0, u, bn_bits(u) - BN_DIGIT);
				_x = t0->dp[0];
				bn_rsh(t1, v, bn_bits(u) - BN_DIGIT);
				_y = t1->dp[0];
				t = 0;
				if (_y != 0) {
					q = _x / _y;
					t = _x % _y;
				}
				if (t >= ((dig_t)1 << BN_DIGIT / 2)) {
					while (1) {
						_q = _y / t;
						_t = _y % t;
						if (_t < ((dig_t)1 << BN_DIGIT / 2)) {
							break;
						}
						_x = _y;
						_y = t;
						t = _a - q * _c;
						_a = _c;
						_c = t;
						t = _b - q * _d;
						_b = _d;
						_d = t;
						t = _t;
						q = _q;
					}
				}
				if (_a < 0) {
					bn_mul_dig(t0, x, -_a);
					bn_neg(t0, t0);
				} else {
					bn_mul_dig(t0, x, _a);
				}
				if (_b < 0) {
					bn_mul_dig(t1, y, -_b);
					bn_neg(t1, t1);
				} else {
					bn_mul_dig(t1, y, _b);
				}
				if (_c < 0) {
					bn_mul_dig(t2, x, -_c);
					bn_neg(t2, t2);
				} else {
					bn_mul_dig(t2, x, _c);
				}
				if (_d < 0) {
					bn_mul_dig(t3, y, -_d);
					bn_neg(t3, t3);
				} else {
					bn_mul_dig(t3, y, _d);
				}
				bn_add(x, t0, t1);
				bn_add(y, t2, t3);
			}
		}
		bn_gcd_ext_dig(c, u, v, x, y->dp[0]);
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(x);
		bn_free(y);
		bn_free(u);
		bn_free(v);
		bn_free(t0);
		bn_free(t1);
		bn_free(t2);
		bn_free(t3);
	}
}

void bn_gcd_ext_lehme(bn_t c, bn_t d, bn_t e, bn_t a, bn_t b) {
	bn_t x, y, u, v, t0, t1, t2, t3, t4;
	dig_t _x, _y, q, _q, t, _t;
	dsi_t _a, _b, _c, _d;
	int swap;

	bn_null(x);
	bn_null(y);
	bn_null(u);
	bn_null(v);
	bn_null(t0);
	bn_null(t1);
	bn_null(t2);
	bn_null(t3);
	bn_null(t4);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		bn_zero(d);
		if (e != NULL) {
			bn_set_dig(e, 1);
		}
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		bn_set_dig(d, 1);
		if (e != NULL) {
			bn_zero(e);
		}
		return;
	}

	if (e == 0 && d && bn_cmp(a, b) == CMP_LT) {
		bn_gcd_ext_gmp(c, d, e, a, b);
		return;
	}

	/*
	 * Taken from Handbook of Hyperelliptic and Elliptic Cryptography.
	 */
	TRY {
		bn_new(x);
		bn_new(y);
		bn_new(u);
		bn_new(v);
		bn_new(t0);
		bn_new(t1);
		bn_new(t2);
		bn_new(t3);
		bn_new(t4);

		if (bn_cmp(a, b) != CMP_LT) {
			bn_abs(x, a);
			bn_abs(y, b);
			swap = 0;
		} else {
			bn_abs(x, b);
			bn_abs(y, a);
			swap = 1;
		}

		bn_zero(t4);
		bn_set_dig(d, 1);

		while (y->used > 1) {
			bn_rsh(u, x, bn_bits(x) - BN_DIGIT);
			_x = u->dp[0];
			bn_rsh(v, y, bn_bits(x) - BN_DIGIT);
			_y = v->dp[0];
			_a = _d = 1;
			_b = _c = 0;
			t = 0;
			if (_y != 0) {
				q = _x / _y;
				t = _x % _y;
			}
			if (t >= ((dig_t)1 << (BN_DIGIT / 2))) {
				while (1) {
					_q = _y / t;
					_t = _y % t;
					if (_t < ((dig_t)1 << (BN_DIGIT / 2))) {
						break;
					}
					_x = _y;
					_y = t;
					t = _a - q * _c;
					_a = _c;
					_c = t;
					t = _b - q * _d;
					_b = _d;
					_d = t;
					t = _t;
					q = _q;
				}
			}
			if (_b == 0) {
				bn_div_rem(t1, t0, x, y);
				bn_copy(x, y);
				bn_copy(y, t0);
				bn_mul(t1, t1, d);
				bn_sub(t1, t4, t1);
				bn_copy(t4, d);
				bn_copy(d, t1);
			} else {
				bn_rsh(u, x, bn_bits(x) - 2 * BN_DIGIT);
				bn_rsh(v, y, bn_bits(x) - 2 * BN_DIGIT);
				if (_a < 0) {
					bn_mul_dig(t0, u, -_a);
					bn_neg(t0, t0);
				} else {
					bn_mul_dig(t0, u, _a);
				}
				if (_b < 0) {
					bn_mul_dig(t1, v, -_b);
					bn_neg(t1, t1);
				} else {
					bn_mul_dig(t1, v, _b);
				}
				if (_c < 0) {
					bn_mul_dig(t2, u, -_c);
					bn_neg(t2, t2);
				} else {
					bn_mul_dig(t2, u, _c);
				}
				if (_d < 0) {
					bn_mul_dig(t3, v, -_d);
					bn_neg(t3, t3);
				} else {
					bn_mul_dig(t3, v, _d);
				}
				bn_add(u, t0, t1);
				bn_add(v, t2, t3);
				bn_rsh(t0, u, bn_bits(u) - BN_DIGIT);
				_x = t0->dp[0];
				bn_rsh(t1, v, bn_bits(u) - BN_DIGIT);
				_y = t1->dp[0];
				t = 0;
				if (_y != 0) {
					q = _x / _y;
					t = _x % _y;
				}
				if (t >= ((dig_t)1 << BN_DIGIT / 2)) {
					while (1) {
						_q = _y / t;
						_t = _y % t;
						if (_t < ((dig_t)1 << BN_DIGIT / 2)) {
							break;
						}
						_x = _y;
						_y = t;
						t = _a - q * _c;
						_a = _c;
						_c = t;
						t = _b - q * _d;
						_b = _d;
						_d = t;
						t = _t;
						q = _q;
					}
				}
				if (_a < 0) {
					bn_mul_dig(t0, x, -_a);
					bn_neg(t0, t0);
				} else {
					bn_mul_dig(t0, x, _a);
				}
				if (_b < 0) {
					bn_mul_dig(t1, y, -_b);
					bn_neg(t1, t1);
				} else {
					bn_mul_dig(t1, y, _b);
				}
				if (_c < 0) {
					bn_mul_dig(t2, x, -_c);
					bn_neg(t2, t2);
				} else {
					bn_mul_dig(t2, x, _c);
				}
				if (_d < 0) {
					bn_mul_dig(t3, y, -_d);
					bn_neg(t3, t3);
				} else {
					bn_mul_dig(t3, y, _d);
				}
				bn_add(x, t0, t1);
				bn_add(y, t2, t3);

				if (_a < 0) {
					bn_mul_dig(t0, t4, -_a);
					bn_neg(t0, t0);
				} else {
					bn_mul_dig(t0, t4, _a);
				}
				if (_b < 0) {
					bn_mul_dig(t1, d, -_b);
					bn_neg(t1, t1);
				} else {
					bn_mul_dig(t1, d, _b);
				}
				if (_c < 0) {
					bn_mul_dig(t2, t4, -_c);
					bn_neg(t2, t2);
				} else {
					bn_mul_dig(t2, t4, _c);
				}
				if (_d < 0) {
					bn_mul_dig(t3, d, -_d);
					bn_neg(t3, t3);
				} else {
					bn_mul_dig(t3, d, _d);
				}
				bn_add(t4, t0, t1);
				bn_add(d, t2, t3);
			}
		}
		bn_gcd_ext_dig(c, u, v, x, y->dp[0]);
		if (!swap) {
			bn_mul(t0, t4, u);
			bn_mul(t1, d, v);
			bn_add(t4, t0, t1);
			bn_mul(x, b, t4);
			bn_sub(x, c, x);
			bn_div(d, x, a);
		} else {
			bn_mul(t0, t4, u);
			bn_mul(t1, d, v);
			bn_add(d, t0, t1);
			bn_mul(x, a, d);
			bn_sub(x, c, x);
			bn_div(t4, x, b);
		}
		if (e != NULL) {
			bn_copy(e, t4);
		}
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(x);
		bn_free(y);
		bn_free(u);
		bn_free(v);
		bn_free(t0);
		bn_free(t1);
		bn_free(t2);
		bn_free(t3);
		bn_free(t4);
	}
}

#endif

#if BN_GCD == STEIN || !defined(STRIP)

void bn_gcd_stein(bn_t c, bn_t a, bn_t b) {
	bn_t u, v, t;
	int shift;

	bn_null(u);
	bn_null(v);
	bn_null(t);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		return;
	}

	TRY {
		bn_new(u);
		bn_new(v);
		bn_new(t);

		bn_abs(u, a);
		bn_abs(v, b);

		shift = 0;
		while (bn_is_even(u) && bn_is_even(v)) {
			bn_hlv(u, u);
			bn_hlv(v, v);
			shift++;
		}
		while (!bn_is_zero(u)) {
			while (bn_is_even(u)) {
				bn_hlv(u, u);
			}
			while (bn_is_even(v)) {
				bn_hlv(v, v);
			}
			bn_sub(t, u, v);
			bn_abs(t, t);
			bn_hlv(t, t);
			if (bn_cmp(u, v) != CMP_LT) {
				bn_copy(u, t);
			} else {
				bn_copy(v, t);
			}
		}
		bn_lsh(c, v, shift);
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(u);
		bn_free(v);
		bn_free(t);
	}
}

void bn_gcd_ext_stein(bn_t c, bn_t d, bn_t e, bn_t a, bn_t b) {
	bn_t x, y, u, v, _a, _b, _e;
	int shift, found;

	bn_null(x);
	bn_null(y);
	bn_null(u);
	bn_null(v);
	bn_null(_a);
	bn_null(_b);
	bn_null(_e);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		bn_zero(d);
		if (e != NULL) {
			bn_set_dig(e, 1);
		}
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		bn_set_dig(d, 1);
		if (e != NULL) {
			bn_zero(e);
		}
		return;
	}

	TRY {
		bn_new(x);
		bn_new(y);
		bn_new(u);
		bn_new(v);
		bn_new(_a);
		bn_new(_b);
		bn_new(_e);

		bn_abs(x, a);
		bn_abs(y, b);

		/* g = 1. */
		shift = 0;
		/* While x and y are both even, x = x/2 and y = y/2, g = 2g. */
		while (bn_is_even(x) && bn_is_even(y)) {
			bn_hlv(x, x);
			bn_hlv(y, y);
			shift++;
		}

		bn_copy(u, x);
		bn_copy(v, y);

		/* u = x, y = v, A = 1, B = 0, C = 0, D = 1. */
		bn_set_dig(_a, 1);
		bn_zero(_b);
		bn_zero(d);
		bn_set_dig(_e, 1);

		if (e != NULL) {
			found = 0;
			while (!found) {
				/* While u is even, u = u/2. */
				while ((u->dp[0] & 0x01) == 0) {
					bn_hlv(u, u);
					/* If A = B = 0 (mod 2) then A = A/2, B = B/2. */
					if ((_a->dp[0] & 0x01) == 0 && (_b->dp[0] & 0x01) == 0) {
						bn_hlv(_a, _a);
						bn_hlv(_b, _b);
					} else {
						/* Otherwise A = (A + y)/2, B = (B - x)/2. */
						bn_add(_a, _a, y);
						bn_hlv(_a, _a);
						bn_sub(_b, _b, x);
						bn_hlv(_b, _b);
					}
				}
				/* While v is even, v = v/2. */
				while ((v->dp[0] & 0x01) == 0) {
					bn_hlv(v, v);
					/* If C = D = 0 (mod 2) then C = C/2, D = D/2. */
					if ((d->dp[0] & 0x01) == 0 && (_e->dp[0] & 0x01) == 0) {
						bn_hlv(d, d);
						bn_hlv(_e, _e);
					} else {
						/* Otherwise C = (C + y)/2, D = (D - x)/2. */
						bn_add(d, d, y);
						bn_hlv(d, d);
						bn_sub(_e, _e, x);
						bn_hlv(_e, _e);
					}
				}
				/* If u >= v then u = u - v, A = A - C, B = B - D. */
				if (bn_cmp(u, v) != CMP_LT) {
					bn_sub(u, u, v);
					bn_sub(_a, _a, d);
					bn_sub(_b, _b, _e);
				} else {
					/* Otherwise, v = v - u, C = C - a, D = D - B. */
					bn_sub(v, v, u);
					bn_sub(d, d, _a);
					bn_sub(_e, _e, _b);
				}
				/* If u = 0 then d = C, e = D and return (d, e, g * v). */
				if (bn_is_zero(u)) {
					bn_lsh(c, v, shift);
					found = 1;
				}
			}
			bn_copy(e, _e);
		} else {
			found = 0;
			while (!found) {
				/* While u is even, u = u/2. */
				while ((u->dp[0] & 0x01) == 0) {
					bn_hlv(u, u);
					/* If A = B = 0 (mod 2) then A = A/2. */
					if ((_a->dp[0] & 0x01) == 0) {
						bn_hlv(_a, _a);
					} else {
						/* Otherwise A = (A + y)/2. */
						bn_add(_a, _a, y);
						bn_hlv(_a, _a);
					}
				}
				/* While v is even, v = v/2. */
				while ((v->dp[0] & 0x01) == 0) {
					bn_hlv(v, v);
					/* If C = D = 0 (mod 2) then C = C/2. */
					if ((d->dp[0] & 0x01) == 0) {
						bn_hlv(d, d);
					} else {
						/* Otherwise C = (C + y)/2. */
						bn_add(d, d, y);
						bn_hlv(d, d);
					}
				}
				/* If u >= v then u = u - v, A = A - C, B = B - D. */
				if (bn_cmp(u, v) != CMP_LT) {
					bn_sub(u, u, v);
					bn_sub(_a, _a, d);
				} else {
					/* Otherwise, v = v - u, C = C - a, D = D - B. */
					bn_sub(v, v, u);
					bn_sub(d, d, _a);
				}
				/* If u = 0 then d = C, e = D and return (d, e, g * v). */
				if (bn_is_zero(u)) {
					bn_lsh(c, v, shift);
					found = 1;
				}
			}
		}
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(x);
		bn_free(y);
		bn_free(u);
		bn_free(v);
		bn_free(_a);
		bn_free(_b);
		bn_free(_e);
	}
}

#endif

void bn_gcd_ext_mid(bn_t c, bn_t d, bn_t e, bn_t f, bn_t a, bn_t b) {
	bn_t q, r, s, t, u, v, x, w, y, z;
	int stop;

	bn_null(q);
	bn_null(r);
	bn_null(s);
	bn_null(t);
	bn_null(u);
	bn_null(v);
	bn_null(x);
	bn_null(w);
	bn_null(y);
	bn_null(z);

	if (bn_is_zero(a)) {
		bn_abs(c, b);
		bn_zero(d);
		bn_zero(e);
		return;
	}

	if (bn_is_zero(b)) {
		bn_abs(c, a);
		bn_set_dig(d, 1);
		bn_set_dig(e, 1);
		return;
	}

	TRY {
		bn_new(q);
		bn_new(r);
		bn_new(s);
		bn_new(t);
		bn_new(u);
		bn_new(v);
		bn_new(x);
		bn_new(w);
		bn_new(y);
		bn_new(z);

		if (bn_cmp_abs(u, v) == CMP_GT) {
			bn_abs(u, a);
			bn_abs(v, b);
		} else {
			bn_abs(u, b);
			bn_abs(v, a);
		}

		stop = bn_bits(u) >> 1;

		bn_set_dig(x, 1);
		bn_zero(t);

		int wait = 0;
		while (!bn_is_zero(v)) {
			bn_div_rem(q, r, u, v);

			bn_copy(u, v);
			bn_copy(v, r);

			bn_mul(s, q, x);
			bn_sub(s, t, s);
			bn_copy(t, x);
			bn_copy(x, s);

			if (wait) {
				bn_copy(e, r);
				bn_neg(f, x);
				wait = 0;
			}
			if (bn_bits(u) > stop) {
				bn_copy(c, r);
				bn_neg(d, x);
				bn_copy(w, u);
				bn_neg(y, t);
				wait = 1;
			}
		}
		/* Compute r as the norm of vector (y, z). */
		bn_sqr(s, w);
		bn_sqr(t, y);
		bn_add(t, t, s);

		/* Compute q as the norm of vector (d, e). */
		bn_sqr(r, e);
		bn_sqr(q, f);
		bn_add(q, q, r);

		/* Output (d, e) as the vector of smaller norm. */
		if (bn_cmp(t, q) == CMP_LT) {
			bn_copy(e, w);
			bn_copy(f, y);
		}
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(q);
		bn_free(r);
		bn_free(s);
		bn_free(t);
		bn_free(u);
		bn_free(v);
		bn_free(x);
		bn_free(w);
		bn_free(y);
		bn_free(z);
	}
}

void bn_gcd_dig(bn_t c, bn_t a, dig_t b) {
	dig_t _u, _v, _t = 0;

	if (bn_is_zero(a)) {
		bn_set_dig(c, b);
		return;
	}

	if (b == 0) {
		bn_abs(c, a);
		return;
	}

	bn_mod_dig(&(c->dp[0]), a, b);
	_v = c->dp[0];
	_u = b;
	while (_v != 0) {
		_t = _v;
		_v = _u % _v;
		_u = _t;
	}
	bn_set_dig(c, _u);
}

void bn_gcd_ext_dig(bn_t c, bn_t d, bn_t e, bn_t a, dig_t b) {
	bn_t u, v, x1, y1, q, r;
	dig_t _v, _q, _t, _u;

	bn_null(u);
	bn_null(v);
	bn_null(x1);
	bn_null(y1);
	bn_null(q);
	bn_null(r);

	if (d == NULL && e == NULL) {
		bn_gcd_dig(c, a, b);
		return;
	}

	if (bn_is_zero(a)) {
		bn_set_dig(c, b);
		if (d != NULL) {
			bn_zero(d);
		}
		if (e != NULL) {
			bn_set_dig(e, 1);
		}
		return;
	}

	if (b == 0) {
		bn_abs(c, a);
		if (d != NULL) {
			bn_set_dig(d, 1);
		}
		if (e != NULL) {
			bn_zero(e);
		}
		return;
	}

	TRY {
		bn_new(u);
		bn_new(v);
		bn_new(x1);
		bn_new(y1);
		bn_new(q);
		bn_new(r);

		bn_abs(u, a);
		bn_set_dig(v, b);

		bn_zero(x1);
		bn_set_dig(y1, 1);
		if (d != NULL) {
			bn_set_dig(d, 1);
		}
		if (e != NULL) {
			bn_zero(e);
		}

		bn_div_rem(q, r, u, v);

		bn_copy(u, v);
		bn_copy(v, r);

		if (d != NULL) {
			bn_mul(c, q, x1);
			bn_sub(r, d, c);
			bn_copy(d, x1);
			bn_copy(x1, r);
		}

		if (e != NULL) {
			bn_mul(c, q, y1);
			bn_sub(r, e, c);
			bn_copy(e, y1);
			bn_copy(y1, r);
		}

		_v = v->dp[0];
		_u = u->dp[0];
		while (_v != 0) {
			_q = _u / _v;
			_t = _u % _v;

			_u = _v;
			_v = _t;

			if (d != NULL) {
				bn_mul_dig(c, x1, _q);
				bn_sub(r, d, c);
				bn_copy(d, x1);
				bn_copy(x1, r);
			}

			if (e != NULL) {
				bn_mul_dig(c, y1, _q);
				bn_sub(r, e, c);
				bn_copy(e, y1);
				bn_copy(y1, r);
			}
		}
		bn_set_dig(c, _u);
	}
	CATCH_ANY {
		THROW(ERR_CAUGHT);
	}
	FINALLY {
		bn_free(u);
		bn_free(v);
		bn_free(x1);
		bn_free(y1);
		bn_free(q);
		bn_free(r);
	}
}

void bn_rsh1_low_ex(dig_t *c, dig_t *a, int size);
void bn_addn_low_ex(dig_t *c, dig_t *a, dig_t *b, int size);
void bn_negn_low_ex(dig_t *c, dig_t *a, int size);
dig_t bn_subn_low_ex(dig_t *c, dig_t *a, dig_t *b, int size);
void bn_ccopyn_low_arm_asm(dig_t *c, dig_t *b, int bit, int digs);

void bn_invmod_low_asm_mem(dig_t *c, dig_t *x, dig_t *m);
void bn_invmod_low_asm_reg(dig_t *c, dig_t *x, dig_t *m);

//#define bn_ccopyn_low bn_ccopyn_low_arm_asm

void bn_invmod(bn_t c, bn_t x, bn_t m) {
#if FB_POLYN == 251 || FB_POLYN == 283
	int i, digs, n;
	n = bn_bits(m);
	digs = n / BN_DIGIT + (n % BN_DIGIT > 0);
	for (i = x->used; i < BN_DIGS; i++) {
		x->dp[i] = 0;
	}
	for (i = m->used; i < BN_DIGS; i++) {
		m->dp[i] = 0;
	}
#if CORTEX == 8 || CORTEX == 9
	bn_invmod_low_asm_mem(c->dp, x->dp, m->dp);
#elif CORTEX == 15
	bn_invmod_low_asm_reg(c->dp, x->dp, m->dp);
#else
#error "CORTEX invalid or not specified"
#endif
	c->used = digs;
#else
	bn_t na, nb, nu, nv, nt1, nt2;
	dig_t *a, *b, *u, *v, *t1, *t2, *tp;
	int i, n, odd, digs, borrow;

	bn_new(na);
	bn_new(nb);
	bn_new(nu);
	bn_new(nv);
	bn_new(nt1);
	bn_new(nt2);

	bn_zero(na);
	bn_zero(nb);
	bn_zero(nu);
	bn_zero(nv);
	bn_zero(nt1);
	bn_zero(nt2);

	n = bn_bits(m);
	bn_copy(na, x);
	bn_set_dig(nu, 1);
	bn_copy(nb, m);
	bn_set_dig(nv, 0);
	bn_add_dig(nt2, m, 1);
	bn_rsh(nt2, nt2, 1);

	a = na->dp;
	b = nb->dp;
	u = nu->dp;
	v = nv->dp;
	t1 = nt1->dp;
	t2 = nt2->dp;

	digs = n / BN_DIGIT + (n % BN_DIGIT > 0);

	for (i = 0; i < 2 * n - 1; i++) {
		odd = a[0] & 1;
		/* a -= odd * b */
		borrow = bn_subn_low_ex(t1, a, b, digs);
		borrow &= odd;
		bn_ccopyn_low(b, a, borrow, digs);
		bn_ccopyn_low(a, t1, odd, digs);
#if 0
		memset(t1, 0, digs * sizeof(dig_t));
		bn_subn_low_ex(t1, t1, a, digs);
#else
		bn_negn_low_ex(t1, a, digs);
#endif
		bn_ccopyn_low(a, t1, borrow, digs);

		bn_rsh1_low_ex(a, a, digs);

		/* if borrow then swap u, v */
		tp = (dig_t *) SELECT((uintptr_t ) u, (uintptr_t ) v, (uintptr_t ) borrow);
		v = (dig_t *) SELECT((uintptr_t ) v, (uintptr_t ) u, (uintptr_t ) borrow);
		u = tp;

		/* u -= odd * v */
		borrow = bn_subn_low_ex(t1, u, v, digs);
		borrow &= odd;
		bn_ccopyn_low(u, t1, odd, digs);
		/* if borrow then u += m */
		bn_addn_low_ex(t1, u, m->dp, digs);
		bn_ccopyn_low(u, t1, borrow, digs);

		odd = u[0] & 1;
		bn_rsh1_low_ex(u, u, digs);

		/* if odd then u += (m + 1) / 2 */
		bn_addn_low_ex(t1, u, t2, digs);
		bn_ccopyn_low(u, t1, odd, digs);
	}
	c->used = digs;
	c->sign = BN_POS;
	memcpy(c->dp, v, digs * sizeof(dig_t));
#endif
}
