/**
 * @file
 *
 * Implementation of side-channel resistant modular inversion.
 *
 * Works only on numbers the same size of the configured binary field.
 *
 * Implemented for 251 and 283 fields. Other fields use a generic implementation,
 * see the higher level code (relic_bn_gcd.c)
 *
 */

#include "relic_fb_low.h"

.arm
.align
.globl bn_rsh1_low_ex
.globl bn_addn_low_ex
.globl bn_subn_low_ex
.globl bn_negn_low_ex
.globl bn_invmod_low_asm_reg
.globl bn_invmod_low_asm_mem
.globl bn_ccopyn_low_arm_asm

/*
 * Some utility functions. The regular bn functiosn can't be used since they
 * do not take constant time (they optimize leading zero digits).
 * The functions below always operate on all digits.
 */

bn_rsh1_low_ex:
#if FB_POLYN == 163
    vldm r1, {d0-d2}
    vshr.u64 q8, q0, #1
    vshr.u64 d18, d2, #1
    vsli.64 d16, d1, #63
    vsli.64 d17, d2, #63
    vstm r0, {d16-d18}
    bx lr
#elif FB_POLYN == 251
    vldm r1, {d0-d3}
    vshr.u64 q8, q0, #1
    vshr.u64 q9, q1, #1
    vsli.64 d16, d1, #63
    vsli.64 d17, d2, #63
    vsli.64 d18, d3, #63
    vstm r0, {d16-d19}
    bx lr
#elif FB_POLYN == 283
#if 1
    vldm r1, {d0-d4}
    vshr.u64 q8, q0, #1
    vshr.u64 q9, q1, #1
    vshr.u64 d20, d4, #1
    vsli.64 d16, d1, #63
    vsli.64 d17, d2, #63
    vsli.64 d18, d3, #63
    vsli.64 d19, d4, #63
    vstm r0, {d16-d20}
    bx lr
#elif 0
    push {r4, r5}
    ldr ip, [r1, #(8*4)]
    ldrd r2, r3, [r1, #(6*4)]
    lsrs ip, ip, #1
    str ip, [r0, #(8*4)]
    rrxs r3, r3
    rrxs r2, r2
    ldrd r4, r5, [r1, #(4*4)]
    strd r2, r3, [r0, #(6*4)]
    rrxs r5, r5
    rrxs r4, r4
    ldrd r2, r3, [r1, #(2*4)]
    strd r4, r5, [r0, #(4*4)]
    rrxs r3, r3
    rrxs r2, r2
    ldrd r4, r5, [r1, #(0*4)]
    strd r2, r3, [r0, #(2*4)]
    rrxs r5, r5
    rrxs r4, r4
    strd r4, r5, [r0, #(0*4)]
    pop {r4, r5}
    bx lr
#else
    push {r4, r5, r6, r7, r8, lr}
    ldm r1, {r2, r3, r4, r5, r6, r7, r8, ip, lr}
    lsrs lr, lr, #1
    rrxs ip, ip
    rrxs r8, r8
    rrxs r7, r7
    rrxs r6, r6
    rrxs r5, r5
    rrxs r4, r4
    rrxs r3, r3
    rrxs r2, r2
    stm r0, {r2, r3, r4, r5, r6, r7, r8, ip, lr}
    pop {r4, r5, r6, r7, r8, pc}
#endif
#elif FB_POLYN == 353
    vldm r1, {d0-d5}
    vshr.u64 q8, q0, #1
    vshr.u64 q9, q1, #1
    vshr.u64 q10, q2, #1
    vsli.64 d16, d1, #63
    vsli.64 d17, d2, #63
    vsli.64 d18, d3, #63
    vsli.64 d19, d4, #63
    vsli.64 d20, d5, #63
    vstm r0, {d16-d21}
    bx lr
#elif FB_POLYN == 571
    vldm r1, {d0-d8}
    vshr.u64 q8, q0, #1
    vshr.u64 q9, q1, #1
    vshr.u64 q10, q2, #1
    vshr.u64 q11, q3, #1
    vshr.u64 d24, d8, #1
    vsli.64 d16, d1, #63
    vsli.64 d17, d2, #63
    vsli.64 d18, d3, #63
    vsli.64 d19, d4, #63
    vsli.64 d20, d5, #63
    vsli.64 d21, d6, #63
    vsli.64 d22, d7, #63
    vsli.64 d23, d8, #63
    vstm r0, {d16-d24}
    bx lr
#endif

bn_addn_low_ex:
#if FB_POLYN == 163
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    adds r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7}
    ldm r1!, {r3, r4}
    adcs r3, r6
    adcs r4, r7
    stm r0!, {r3, r4}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 251
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    adds r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 283
#if 1
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    adds r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldr r6, [r2]
    ldr r3, [r1]
    adc r3, r6
    str r3, [r0]
    pop {r4, r5, r6, r7, r8, pc}
#else
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4}
    ldm r2!, {r6, r7}
    adds r3, r6
    adcs r4, r7
        ldm r1!, {r5, ip}
        ldm r2!, {r8, lr}
    stm r0!, {r3, r4}
        adcs r5, r8
        adcs ip, lr
    ldm r1!, {r3, r4}
    ldm r2!, {r6, r7}
        stm r0!, {r5, ip}
    adcs r3, r6
    adcs r4, r7
        ldm r1!, {r5, ip}
        ldm r2!, {r8, lr}
    stm r0!, {r3, r4}
        adcs r5, r8
        adcs ip, lr
    ldr r6, [r2]
    ldr r3, [r1]
        stm r0!, {r5, ip}
    adc r3, r6
    str r3, [r0]
    pop {r4, r5, r6, r7, r8, pc}
#endif
#elif FB_POLYN == 353
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    adds r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 571
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    adds r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    adcs r3, r6
    adcs r4, r7
    adcs r5, r8
    adcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7}
    ldm r1!, {r3, r4}
    adcs r3, r6
    adcs r4, r7
    stm r0!, {r3, r4}
    pop {r4, r5, r6, r7, r8, pc}
#endif

bn_subn_low_ex:
#if FB_POLYN == 163
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7}
    ldm r1!, {r3, r4}
    sbcs r3, r6
    sbcs r4, r7
    stm r0!, {r3, r4}
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 251
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 283
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldr r6, [r2]
    ldr r3, [r1]
    sbcs r3, r6
    str r3, [r0]
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 353
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 571
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, lr
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7}
    ldm r1!, {r3, r4}
    sbcs r3, r6
    sbcs r4, r7
    stm r0!, {r3, r4}
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, pc}
#endif

bn_negn_low_ex:
#if FB_POLYN == 163
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    rsbs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4}
    rscs r3, #0
    rscs r4, #0
    stm r0!, {r3, r4}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 251
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    rsbs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 283
    push {r4, r5}
    ldm r1!, {r3, r4, r5, ip}
    rsbs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldr r3, [r1]
    rscs r3, #0
    str r3, [r0]
    pop {r4, r5}
    bx lr
#elif FB_POLYN == 353
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    rsbs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    pop {r4, r5, r6, r7, r8, pc}
#elif FB_POLYN == 571
    push {r4, r5, r6, r7, r8, lr}
    ldm r1!, {r3, r4, r5, ip}
    rsbs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4, r5, ip}
    rscs r3, #0
    rscs r4, #0
    rscs r5, #0
    rscs ip, #0
    stm r0!, {r3, r4, r5, ip}
    ldm r1!, {r3, r4}
    rscs r3, #0
    rscs r4, #0
    stm r0!, {r3, r4}
    pop {r4, r5, r6, r7, r8, pc}
#endif

#if 0
    push {r4, r5, r6, r7, r8, r9}
    ldm r1!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, r9}
    subs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, r9
    stm r0!, {r3, r4, r5, ip}
    ldm r2!, {r6, r7, r8, r9}
    ldm r1!, {r3, r4, r5, ip}
    sbcs r3, r6
    sbcs r4, r7
    sbcs r5, r8
    sbcs ip, r9
    stm r0!, {r3, r4, r5, ip}
    ldr r6, [r2]
    ldr r3, [r1]
    sbcs r3, r6
    str r3, [r0]
    mov r0, #0
    sbc r0, #0
    rsb r0, #0
    pop {r4, r5, r6, r7, r8, r9}
    bx lr


    vldm r1, {d16-d20}
    vldm r2, {d26-d30}

    vaddl.u32 q0, d16, d26
    vaddl.u32 q1, d17, d27
    vaddl.u32 q2, d18, d28
    vaddl.u32 q3, d19, d29
    vaddl.u32 q15, d20, d30

    vshr.u64 d21, d0, #32
    vadd.i64 d1, d1, d21
    vshr.u64 d21, d1, #32
    vadd.i64 d2, d2, d21
    vshr.u64 d21, d2, #32
    vadd.i64 d3, d3, d21
    vshr.u64 d21, d3, #32
    vadd.i64 d4, d4, d21


    vst1.64 {d0}, [r0:64]!
    vaddl.u32 q0, d17, d1
#endif

/* The actual inversion, see the included files. */

#if FB_POLYN == 251
#include "relic_bn_util_low_251.s"
#elif FB_POLYN == 283
#include "relic_bn_util_low_283.s"
#endif
